/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * This is GNU GO, a Go program. Contact gnugo@gnu.org, or see   *
 * http://www.gnu.org/software/gnugo/ for more information.      *
 *                                                               *
 * Copyright 1999 and 2000 by the Free Software Foundation.      *
 *                                                               *
 * This program is free software; you can redistribute it and/or *
 * modify it under the terms of the GNU General Public License   *
 * as published by the Free Software Foundation - version 2.     *
 *                                                               *
 * This program is distributed in the hope that it will be       *
 * useful, but WITHOUT ANY WARRANTY; without even the implied    *
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR       *
 * PURPOSE.  See the GNU General Public License in file COPYING  *
 * for more details.                                             *
 *                                                               *
 * You should have received a copy of the GNU General Public     *
 * License along with this program; if not, write to the Free    *
 * Software Foundation, Inc., 59 Temple Place - Suite 330,       *
 * Boston, MA 02111, USA                                         *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


/*
 * This file is about life & death.
 * 
 * For now, we only consider eye spaces.  Code for extending and
 * shrinking eye spaces is in mouse.c. 
 *
 * We will only handle small eye spaces for now (<= 6 during testing).
 * At these small sizes, we make the simplifying assumption that
 * if any part of the surrounding wall is captured, the resulting
 * eye space will lead to 0 eyes.  This assumtion can be dealt with
 * later when the basic code works.
 *
 * Strategy:
 * Try to fill the entire eye space until there are only eyes of
 * size 1 left. Then nobody can play there, and we can count the eyes.
 */

/* FIXME: TODO
 *  - Handle the case that the wall is put into atari.
 *  - Handle ko
 *  - Handle seki
 *  - Sometimes the correct move for the defending side is not the same
 *    as the attacking move.  Or there might be only one defending move 
 *    but several attacking moves.  The following is such a case:
 *
 *      XXXXX
 *      XOOOXXXXX
 *      X..OOOOOX
 *      ..........
 *      ----------
 *
 *      XXXXX
 *      XOOOXXXXX
 *      Xa.OOOOOX
 *      ..bcdef..
 *      ---------
 *
 *    O can play at a and make two eyes, but X can play at any of
 *    a to f and reduce the number of eyes to 1. Another case is this:
 *
 *      |XXX..
 *      |OOX..
 *      |.OXXX
 *      |XOO.X
 *      |.XO.X
 *      +------
 *
 *      |XXX..
 *      |OOX..
 *      |aOXXX
 *      |XOO.X
 *      |bXO.X
 *      +------
 *    where O must play at a to live unconditionally, but X must play at b
 *    to kill unconditionally.  In this case, there is no overlap between
 *    the correct play for the attacking and the defending side.
 *
 *    So we should make recognize_eye return both a defending move and an 
 *    attacking move.
 * - Try to find out how many outside liberties are needed for the optimal
 *   sequence.
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "liberty.h"
#include "location_set.h"


#define MAX_EYE_SIZE  10

static int minimize_eyes(int color, int *min, int *attacki, int *attackj, 
			  Location_set *eye, int last_move_was_pass);
static int maximize_eyes(int color, int *max, int *attacki, int *attackj, 
			 Location_set *eye, int last_move_was_pass);
static void life_showboard(void);

/* Used by life_showboard(). */
static int stackp_when_called = 0;

/* ================================================================ */
/*                   Hash table for recognize_eye.                  */
/* ================================================================ */


/*
 * Layout of the hash table:
 *
 *   Each entry is one 32 bit integer. (FIXME: Don't waste 32 bits on 
 *   64 bit machines.) 
 *
 *   byte 0 (LSB):  Result byte (see below)
 *   byte 1:        Where the attacker should play (index in the set)
 *   byte 2:        Where the attacker should play.
 *   byte 3 (MSB):  Reserved
 *
 *   The result byte:
 *   +-------------------------------+
 *   | 7 | 6 | 5 | 4 | 3 | 2 | 1 | 0 |
 *   +-------------------------------+
 *   |vis|xxx|  dresult  |  aresult  |
 *   +-------------------------------+
 *
 *   aresult and dresult:
 *   0:   No result known yet.  This is used during the search.
 *   1:   0 eyes
 *   2:   1 eye
 *   3:   2 eyes
 *   4-7: Reserved
 */

/* defines for the entire entry */
#define ATTACK_MOVE_SHIFT  8
#define ATTACK_BYTE        (0xff << ATTACK_MOVE_SHIFT) /* bit mask */
#define ATTACK_NO_MOVE     (255 << ATTACK_MOVE_SHIFT)  /* value 255 */
#define DEFENSE_MOVE_SHIFT 16
#define DEFENSE_BYTE       (0xff << DEFENSE_MOVE_SHIFT) /* bit mask */
#define DEFENSE_NO_MOVE    (255 << DEFENSE_MOVE_SHIFT)

/* defines for the result byte */
#define VISITED              0x80	/* Set if this node is visited */
#define RESULT_MASK          0x07
#define ATTACK_RESULT_SHIFT  0
#define DEFENSE_RESULT_SHIFT 3


/* The hash table will not have to handle any collisions, since it
 * is complete. 
 */
#define EYE_HASH_SIZE  3*3*3*3*3*3*3*3*3*3*2 /* 3^MAX_EYE_SIZE*2 */


static int  eyehash_table[EYE_HASH_SIZE];
#if 0 /* Unused /gf */
static int  eyehash_initialized = 0;
#endif

static void  eyehash_clear(void);
static int   eyehash_index(Location_set *eyedata, int to_move);


static void
eyehash_clear()
{
  int  i;

  for (i = 0; i < EYE_HASH_SIZE; ++i)
    eyehash_table[i] = 0;
}


static int
eyehash_index(Location_set *eyepoints, int to_move)
{
  Location_set_iterator   iter;
  int  hashval;
  int  i, j;

  hashval = 0;
  for (ls_iter_init(eyepoints, &iter);
       ls_iter_next(&iter, &i, &j);
       ls_iter_update(&iter))
    hashval = hashval * 3 + p[i][j];
  hashval = hashval * 2 + (to_move - 1);

  return hashval;
}



static void
eyehash_get_attack_move(Location_set *eyepoints, int hashindex,
			int *attacki, int *attackj)
{
  Location_set_iterator  iter;
  int  move_index;
  int  m, n;
  int  i;

  move_index = (eyehash_table[hashindex] >> ATTACK_MOVE_SHIFT) & 255;

  for (ls_iter_init(eyepoints, &iter), i = 0;
       ls_iter_next(&iter, &m, &n), i < move_index;
       ls_iter_update(&iter), i++)
    /* NOTHING */;

  if (attacki) *attacki = m;
  if (attackj) *attackj = n;
}


static void
eyehash_set_attack_move(Location_set *eyepoints, int hashindex,
			int attacki, int attackj)
{
  Location_set_iterator  iter;
  int  m,n;
  int  i;

  for (ls_iter_init(eyepoints, &iter), i = 0;
       ls_iter_next(&iter, &m, &n);
       ls_iter_update(&iter), i++)
    if (m == attacki && n == attackj) {
      eyehash_table[hashindex] |= i << ATTACK_MOVE_SHIFT;
      break;
    }
}


static void
eyehash_get_defense_move(Location_set *eyepoints, int hashindex,
			 int *defensei, int *defensej)
{
  Location_set_iterator  iter;
  int  move_index;
  int  m, n;
  int  i;

  move_index = (eyehash_table[hashindex] >> DEFENSE_MOVE_SHIFT) & 255;

  for (ls_iter_init(eyepoints, &iter), i = 0;
       ls_iter_next(&iter, &m, &n), i < move_index;
       ls_iter_update(&iter), i++)
    /* NOTHING */;

  if (defensei) *defensei = m;
  if (defensej) *defensej = n;
}


static void
eyehash_set_defense_move(Location_set *eyepoints, int hashindex,
			 int defensei, int defensej)
{
  Location_set_iterator  iter;
  int  m,n;
  int  i;

  for (ls_iter_init(eyepoints, &iter), i = 0;
       ls_iter_next(&iter, &m, &n);
       ls_iter_update(&iter), i++)
    if (m == defensei && n == defensej) {
      eyehash_table[hashindex] |= i << DEFENSE_MOVE_SHIFT;
      break;
    }
}


/* ================================================================ */
/*                          Eye recognition                         */
/* ================================================================ */


/* The boundary of the eye we are investigating. */
static Location_set  boundary;


/*
 * Compute the maximum and minimum number of eyes reachable from the
 * eye space at (m, n). The maximum is defined for color to play first,
 * and the minimum is defined for other_color to play first.
 * If max != min, (*attacki, *attackj) is set to the move which
 * reduces the eye. The eye data is taken from eye.
 */

void
recognize_eyes2(int m, int  n, 
		int *attacki, int *attackj, int *defensei, int *defensej,
		int *max, int *min, 
		struct eye_data eyedata[MAX_BOARD][MAX_BOARD])
{
  int color;
  int i, j;
  Location_set           eyepoints;
  Location_set_iterator  iter;
  int   result1, result2;
  int   min1, max1;
  int   attacki1, attackj1;
  int   defensei1, defensej1;
  int  save_stackp;

  save_stackp = stackp;

  /* For use by debug outputs. */
  stackp_when_called = stackp;
  
  /* Set `color' to the owner of the eye. */
  color = eyedata[m][n].color;
  if (color == BLACK_BORDER)
    color = BLACK;
  if (color == WHITE_BORDER)
    color = WHITE;

  /* Collect all the points of the eye into the set. */
  location_set_clear(&eyepoints);
  for (i = 0; i < board_size; ++i)
    for (j = 0; j < board_size; ++j) {
      if (eyedata[i][j].origini == m && eyedata[i][j].originj == n) 
	location_set_add(&eyepoints, i, j);
    }

  /* Get the boundary. */
  location_set_copy(&boundary, &eyepoints);
  location_set_expand(&boundary, 0);
  location_set_subtract(&boundary, &eyepoints);

  /* Only keep those parts of the boundary that are actually there. */
  for (ls_iter_init(&boundary, &iter);
       ls_iter_next(&iter, &i, &j);
       ls_iter_update(&iter))
  {
    if (p[i][j] != color)
      location_set_del(&boundary, i, j);
  }

  /* Clear the hash table for this eye. */
  eyehash_clear();

  /*  fprintf(stderr, "Eye (%s): ", color == BLACK ? "black" : "white");
  location_set_print(stderr, &eyepoints);*/

  result1 = minimize_eyes(color, &min1, &attacki1, &attackj1, &eyepoints, 0);
  result2 = maximize_eyes(color, &max1, &defensei1, &defensej1, &eyepoints, 0);

  *min = min1;
  *max = max1;
  if (min1 != max1) {
    if (attacki) *attacki = attacki1;
    if (attackj) *attackj = attackj1;
    if (defensei) *defensei = defensei1;
    if (defensej) *defensej = defensej1;
  }

  assert(stackp == save_stackp);
}


/*
 * We have a set of locations in eyepoints forming an eye shape.
 * Check max and min number of eyes we can get if the opponent moves 
 * first.
 *
 * Return values:
 *   0: Ok. Min eyes in *min.
 *  -1: No result.  A cycle in the reading.
 */

static int
minimize_eyes(int color, int *min, int *attacki, int *attackj, 
	      Location_set *eyepoints, int last_move_was_pass)
{
  int  other = OTHER_COLOR(color);
  int  i, j;
  int  m, n;
  Location_set_iterator   iter;
  Location_set_iterator   iter2;
  int  num_other;
  int  num_legal_moves;
  int  num_eyes;
  int  bordercolor;
  int  eyesize;
  int  max1;
  int  localmax;
  int  hashindex;
  int  result;
  int  attacki1, attackj1;
  int  localattacki = -1;
  int  localattackj = -1;

  int  save_stackp;

  TRACE("entering minimize_eyes: stackp = %d\n", stackp);
  if (verbose)
    life_showboard();

  /* Collect how many moves the opposite color can do. */
  num_other = 0;
  num_legal_moves = 0;
  num_eyes = 0;
  for (ls_iter_init(eyepoints, &iter);
       ls_iter_next(&iter, &i, &j);
       ls_iter_update(&iter))
  {
    if (p[i][j] == color)
      continue;

    if (p[i][j] == other) {
      num_other++;
      continue;
    } 

    if (legal(i, j, other))
      num_legal_moves++;
    else {
      bordercolor = find_border(i, j, NULL, &eyesize);
      if (eyesize == 1)
	num_eyes++;
    } 
  }

  /* Check the hash table and see if we have been here before. */
  /* FIXME: This could be calculated incrementally.  Later. */
  hashindex = eyehash_index(eyepoints, other);
  if (eyehash_table[hashindex] != 0) {
    result = (eyehash_table[hashindex] >> ATTACK_RESULT_SHIFT) & RESULT_MASK;
    if (result == 0) {
      /* No result known yet.  We have come back to the same position
	 again.  Just return and hope that we get a better result on
	 some other branch. */
      TRACE("exiting minimize_eyes\n");
      return -1;
    }
    else if (result <= 3) {
      *min = result - 1;
      if ((eyehash_table[hashindex] & ATTACK_BYTE) != ATTACK_NO_MOVE)
	eyehash_get_attack_move(eyepoints, hashindex, attacki, attackj);

      TRACE("exiting minimize_eyes\n");
      return 0;
    }
  }

  /* Mark this node as visited, but without result so far. */
  eyehash_table[hashindex] |= VISITED;

  /* Check if we are done. This means that there are no stones of
   * the attacking color and that all the eye points are illegal
   * to move into. The only way this could happen is that they are
   * of size 1 and without possibility to capture anything.
   */
  if (num_other == 0 && num_legal_moves == 0) {
    if (num_eyes > 2)
      num_eyes = 2;
    *min = num_eyes;
    eyehash_table[hashindex] |= ATTACK_NO_MOVE;
    eyehash_table[hashindex] |= (num_eyes + 1) << ATTACK_RESULT_SHIFT;

    TRACE("exiting minimize_eyes\n");
    return 0;
  } 

  /* Start by see if doing nothing is enough. */
  localmax = 9999;
  if (!last_move_was_pass) {
    result = maximize_eyes(color, &max1, &attacki1, &attackj1, eyepoints, 1);
    if (result == 0) {
      if (max1 < localmax) {
	localmax = max1;
	localattacki = -1;
	localattackj = -1;
      }

      /* Check if the pass yielded 0 eyes.  If so, we can as well
       * return here. */
      if (localmax == 0) {
	*min = 0;
	if (attacki) *attacki = -1;
	if (attackj) *attackj = -1;
	eyehash_table[hashindex] |= ATTACK_NO_MOVE;
	eyehash_table[hashindex] |= (0 + 1) << ATTACK_RESULT_SHIFT;

	TRACE("exiting minimize_eyes\n");
	return 0;
      }
    }  
  }

  /* Now try all possible moves and see what we get. */
  for (ls_iter_init(eyepoints, &iter);
       ls_iter_next(&iter, &i, &j);
       ls_iter_update(&iter))
  {
    int  boundary_captured;

    if (p[i][j] == color)
      continue;

    if (p[i][j] == other) {
      num_other++;
      continue;
    } 

    /* FIXME: Check here if any parts of the walls are captured.  If so,
     *        we return and tell the caller that 0 eyes can be made. */
    save_stackp = stackp;
    boundary_captured = 0;
    TRACE("minimize_eyes: trymove at %m\n", i, j);
    if (trymove(i, j, other, "minimize_eyes", -1, -1)) {
      num_legal_moves++;

      /* Check if any part of the boundary got captured.  If so, we assume
       * that we can get no eyes.
       */
      for (ls_iter_init(&boundary, &iter2);
	   ls_iter_next(&iter2, &m, &n);
	   ls_iter_update(&iter2))
      {
	if (p[m][n] == EMPTY) {
	  boundary_captured = 1;
	  break;
	}
      }

      if (!boundary_captured) {
	/* No boundary string was captured. The opponent has made his move.
	 * Now let's answer him and see how many eyes we can get.
	 */
	result = maximize_eyes(color, &max1, &attacki1, &attackj1, 
			       eyepoints, 0);
	if (result == 0) {
	  if (max1 < localmax) {
	    localmax = max1;
	    localattacki = i;
	    localattackj = j;
	  }
	}
      }

      popgo();
    }
    assert(stackp == save_stackp);

    /* Check if the boundary was captured or the move yielded 0 eyes.
     * If so, we can as well return here. */
    if (boundary_captured || localmax == 0) {
      *min = 0;
      if (attacki) *attacki = i;
      if (attackj) *attackj = j;

      eyehash_set_attack_move(eyepoints, hashindex, i, j);
      eyehash_table[hashindex] |= (0 + 1) << ATTACK_RESULT_SHIFT;

      TRACE("exiting minimize_eyes\n");
      return 0;
    }
  }

  /* If we didn't find anywhere to play, see how many eye spaces there were. */
  if (localmax == 9999) {
    localmax = num_eyes;
    if (localmax > 2)
      localmax = 2;
    *min = localmax;
    if (attacki) *attacki = -1;
    if (attackj) *attackj = -1;

    eyehash_table[hashindex] |= ATTACK_NO_MOVE;
    eyehash_table[hashindex] |= (localmax + 1) << ATTACK_RESULT_SHIFT;

    TRACE("exiting minimize_eyes\n");
    return 0;
  }

  if (localmax > 2)
    localmax = 2;

  *min = localmax;
  if (attacki) *attacki = localattacki;
  if (attackj) *attackj = localattackj;

  eyehash_set_attack_move(eyepoints, hashindex, localattacki, localattackj);
  eyehash_table[hashindex] |= (localmax + 1) << ATTACK_RESULT_SHIFT;

  TRACE("exiting minimize_eyes\n");
  return 0;
}


static int
maximize_eyes(int color, int *max, int *defensei, int *defensej, 
	      Location_set *eyepoints, int last_move_was_pass)
{
  int  other = OTHER_COLOR(color);
  int  i, j;
  Location_set_iterator   iter;
  int  bordercolor;
  int  eyesize;
  int  min1;
  int  localmin;
  int  result;
  int  hashindex;
  int  defensei1, defensej1;
  int  localdefensei = -1;
  int  localdefensej = -1;

  int  save_stackp;

  TRACE("entering maximize_eyes: stackp = %d\n", stackp);
  if (verbose)
    life_showboard();
  localmin = 0;

  /* Check the hash table and see if we have been here before. */
  /* FIXME: This could be calculated incrementally.  Later. */
  hashindex = eyehash_index(eyepoints, color);
  if (eyehash_table[hashindex] != 0) {
    result = (eyehash_table[hashindex] >> DEFENSE_RESULT_SHIFT) & RESULT_MASK;
    if (result == 0) {
      /* No result known yet.  We have come back to the same position
	 again.  Just return and hope that we get a better result on
	 some other branch. */

      TRACE("exiting maximize_eyes\n");
      return -1;
    }
    else if (result <= 3) {
      *max = result - 1;
      if ((eyehash_table[hashindex] & DEFENSE_BYTE) != DEFENSE_NO_MOVE)
	eyehash_get_defense_move(eyepoints, hashindex, defensei, defensej);

      TRACE("exiting maximize_eyes\n");
      return 0;
    }
  }

  /* Mark this node as visited, but without result so far. */
  eyehash_table[hashindex] |= VISITED;

  /* Start by see if doing nothing is enough. */
  if (!last_move_was_pass) {
    result = minimize_eyes(color, &min1, &defensei1, &defensej1, eyepoints, 1);
    if (result == 0) {
      if (min1 > localmin) {
	localmin = min1;
	localdefensei = -1;
	localdefensej = -1;
      }
    }
  }

  for (ls_iter_init(eyepoints, &iter);
       ls_iter_next(&iter, &i, &j);
       ls_iter_update(&iter))
  {
    if (p[i][j] == color)
      continue;

    if (p[i][j] == other)
      continue;

    /* If the eye is of size 1, and none of the neighbouring strings
       is in atari, then this is a proper eye. Don't play there. */
    bordercolor = find_border(i, j, NULL, &eyesize);
    if (eyesize == 1) {
      if ((i == 0 
	   || (p[i-1][j] == color
	       && approxlib(i-1, j, color, 2, NULL, NULL) > 1))
	  && (i == board_size-1
	      || (p[i+1][j] == color
		  && approxlib(i+1, j, color, 2, NULL, NULL) > 1))
	  && (j == 0
	      || (p[i][j-1] == color
		  && approxlib(i, j-1, color, 2, NULL, NULL) > 1))
	  && (j == board_size-1
	      || (p[i][j+1] == color
		  && approxlib(i, j+1, color, 2, NULL, NULL) > 1)))
	continue;
    }

    save_stackp = stackp;
    TRACE("maximize_eyes: trymove at %m\n", i, j);
    if (trymove(i, j, color, "maximize_eyes", -1, -1)) {

      /* Ok, we made our move.  Now let the opponent do his,
       *     and see how many eyes we can get. */
      result = minimize_eyes(color, &min1, &defensei1, &defensej1,
			     eyepoints, 0);
      if (result == 0) {
	if (min1 > localmin) {
	  localmin = min1;
	  localdefensei = i;
	  localdefensej = j;
	}
      }

      popgo();
    }
    assert(stackp == save_stackp);
  }
  
  if (localmin > 2)
    localmin = 2;

  *max = localmin;
  if (defensei) *defensei = localdefensei;
  if (defensej) *defensej = localdefensej;

  if (localdefensei == -1)
    eyehash_table[hashindex] |= ATTACK_NO_MOVE;
  else
    eyehash_set_defense_move(eyepoints, hashindex, 
			     localdefensei, localdefensej);

    eyehash_table[hashindex] |= (localmin + 1) << DEFENSE_RESULT_SHIFT;

  TRACE("exiting maximize_eyes\n");
  return 0;
}

static void
life_showboard()
{
  int i, j;
  start_draw_board();
  for (i=0; i<board_size; i++) {
    for (j=0; j<board_size; j++) {
      int c;
      int color = p[i][j];
      if (color == WHITE) {
	if (move_in_stack(i, j, stackp_when_called))
	  c = 'o';
	else
	  c = 'O';
      }
      else if (color == BLACK) {
	if (move_in_stack(i, j, stackp_when_called))
	  c = 'x';
	else
	  c = 'X';
      }
      else
	c = EMPTY;
      draw_char(i, j, c);
    }
  }
  end_draw_board();
}

/*
 * Local Variables:
 * tab-width: 8
 * c-basic-offset: 2
 * End:
 */
